#include <xsTypes.h>
#include <stlToCArrays.h>
#include <functionMap.h>
#include <XSUtil/Utils/XSutility.h>
#include <cmath>
#include <cfortran.h>

// Functions called FROM here to Fortran:

PROTOCCALLSFSUB19(DOCMPS,docmps,FLOATV,INT,FLOATV,INT,FLOATV,DOUBLEV,DOUBLEV,DOUBLEV,DOUBLEV,FLOATV,
        DOUBLEV,DOUBLEV,DOUBLEV,DOUBLEV,DOUBLEV,DOUBLEV,INTV,INT,INT)
#define DOCMPS(ear,ne,param,Ifl,photar,phcon,phref,phblb,obj,phtot,obscon,obsblb,obsref,obscs,obsbba,objold,iob,ne1,nc) \
               CCALLSFSUB19(DOCMPS,docmps,FLOATV,INT,FLOATV,INT,FLOATV,DOUBLEV,DOUBLEV,DOUBLEV,DOUBLEV,FLOATV, \
               DOUBLEV,DOUBLEV,DOUBLEV,DOUBLEV,DOUBLEV,DOUBLEV,INTV,INT,INT, \
               ear,ne,param,Ifl,photar,phcon,phref,phblb,obj,phtot,obscon,obsblb,obsref,obscs,obsbba,objold,iob,ne1,nc)

PROTOCCALLSFSUB6(QSPLINT,qsplint,DOUBLEV,DOUBLEV,DOUBLEV,INT,DOUBLE,PDOUBLE)
#define QSPLINT(xa,ya,y2a,n,x,y) \
             CCALLSFSUB6(QSPLINT,qsplint,DOUBLEV,DOUBLEV,DOUBLEV,INT,DOUBLE,PDOUBLE, \
             xa,ya,y2a,n,x,y)

PROTOCCALLSFSUB18(MZREFLIMP,mzreflimp,DOUBLEV,DOUBLEV,DOUBLEV,DOUBLEV,DOUBLEV,FLOATV,INT,INT,INT,INT,INT,INT,INT,
         FLOATV,FLOATV,DOUBLEV,DOUBLEV,DOUBLEV)
#define MZREFLIMP(x,compsp,xr,refl,angle,par_refl,nrefl,jmax,jref,nc,kk,ioniz,irefl,spref,xref,xlog,dspxx,f2) \
             CCALLSFSUB18(MZREFLIMP,mzreflimp,DOUBLEV,DOUBLEV,DOUBLEV,DOUBLEV,DOUBLEV,FLOATV,INT,INT,INT,INT,INT,INT,INT, \
             FLOATV,FLOATV,DOUBLEV,DOUBLEV,DOUBLEV,x,compsp,xr,refl,angle,par_refl,nrefl,jmax,jref,nc,kk,ioniz,irefl, \
             spref,xref,xlog,dspxx,f2)


// Functions called from Fortran TO here:

void MZrefl(double *x, double *compsp, double *xr, double *refl, double *angle,
        float *par_refl, int nrefl, int jmax, int jref, int nc, int kk, int ioniz, int irefl);
FCALLSCSUB13(MZrefl,MZREFL,mzrefl,DOUBLEV,DOUBLEV,DOUBLEV,DOUBLEV,DOUBLEV,FLOATV,
        INT,INT,INT,INT,INT,INT,INT)
          

float spco(const float y, const float gamma, const float ec);
FCALLSCFUN3(FLOAT,spco,SPCO,spco,FLOAT,FLOAT,FLOAT)


namespace {
   void comppsCommon(double **pXlog, double **pDspxx, double **pF2, int *pSz);
}

void xscompps(const RealArray& energyArray, const RealArray& params,
        int spectrumNumber, RealArray& flux, RealArray& fluxErr, 
        const string& initString)
{
   using namespace XSutility;
   
   const int NC = 5; // number of points in Gauss quadrature (used in angular integrals)
   
   const int nE = static_cast<int>(energyArray.size()) - 1;
   const int nE1 = nE + 1;
   
   auto_array_ptr<double> apPhcon(new double[nE1]);
   auto_array_ptr<double> apPhref(new double[nE1]);
   auto_array_ptr<double> apPhblb(new double[nE1]);
   auto_array_ptr<double> apObj(new double[nE1]);
   auto_array_ptr<float> apPhtot(new float[nE1]);
   auto_array_ptr<double> apObscon(new double[nE1]);
   auto_array_ptr<double> apObsblb(new double[nE1]);
   auto_array_ptr<double> apObsref(new double[nE1]);
   auto_array_ptr<double> apObscs(new double[nE1*NC]);
   auto_array_ptr<double> apObsbba(new double[nE1*NC]);
   auto_array_ptr<double> apObjold(new double[nE1]);
   auto_array_ptr<int> apIob(new int[nE1]);
   
   float *ear=0, *pars=0, *photar=0, *photer=0;
   XSFunctions::stlToFloatArrays(energyArray, params, flux, fluxErr,
           ear, pars, photar, photer);
   auto_array_ptr<float> apEar(ear);
   auto_array_ptr<float> apPars(pars);
   auto_array_ptr<float> apPhotar(photar);
   auto_array_ptr<float> apPhoter(photer);
   
   DOCMPS(ear, nE, pars, spectrumNumber, photar, apPhcon.get(), apPhref.get(),
        apPhblb.get(), apObj.get(), apPhtot.get(), apObscon.get(), apObsblb.get(),
        apObsref.get(), apObscs.get(), apObsbba.get(), apObjold.get(), apIob.get(),
        nE1, NC);
   
   XSFunctions::floatFluxToStl<float>(photar, photer, nE, false, flux, fluxErr);
}


float spco(const float y, const float gamma, const float ec)
{
   // compute n(1/lambda)*lambda^(-2) = F(1/lambda)/lambda
   
   float val=0.0;
   double *xlog=0;
   double *dspxx=0;
   double *f2=0;
   int sz=0;
   
   // get previously allocated memory for arrays
   comppsCommon(&xlog, &dspxx, &f2, &sz);
   
   const double x = log(static_cast<double>(1.0/y));
   double yyy = 0.0;
   // use of the spline computed
   QSPLINT(xlog, dspxx, f2, sz, x, yyy);
   val = exp(static_cast<float>(yyy))/y;
   
   return val;
}

void MZrefl(double *x, double *compsp, double *xr, double *refl, double *angle,
        float *par_refl, int nrefl, int jmax, int jref, int nc, int kk, int ioniz, int irefl)
{
   // Wrapper function for allocating and passing memory to MZreflImp in compps.f
   
   using namespace XSutility;
   
   static int jrefsav = 0;
   static auto_array_ptr<float> apSpref(0);
   static auto_array_ptr<float> apXref(0);   
   if (jref != jrefsav)
   {
      float *oldF = apSpref.reset(new float[jref]);
      delete [] oldF;
      oldF = apXref.reset(new float[jref]);
      delete [] oldF;
      jrefsav = jref;
   }
   float *spref = apSpref.get();
   float *xref = apXref.get();
   
   double *xlog=0;
   double *dspxx=0;
   double *f2=0;
   // Even though we're passing its pointer, jmax's value is never
   // changed in this context since presumably it's never 0 in here.
   comppsCommon(&xlog, &dspxx, &f2, &jmax);
   
   MZREFLIMP(x, compsp, xr, refl, angle, par_refl, nrefl, jmax, jref, nc, 
            kk, ioniz, irefl, spref, xref, xlog, dspxx, f2);
}



namespace {
void comppsCommon(double **pXlog, double **pDspxx, double **pF2, int *pSz)
{
   using namespace XSutility;
   static auto_array_ptr<double> apXlog(0);
   static auto_array_ptr<double> apDspxx(0);
   static auto_array_ptr<double> apF2(0);
   static int sz = 0;
   
   // If size indicated by *pSz is new (and non-zero), allocate new 
   // memory.  Otherwise just return pointers to existing memory and
   // the size of the arrays.  
   
   if (*pSz != 0 && *pSz != sz)
   {
      sz = *pSz;
      double *oldD = apXlog.reset(new double[sz]);
      delete [] oldD;
      oldD = apDspxx.reset(new double[sz]);
      delete [] oldD;
      oldD = apF2.reset(new double[sz]);
      delete [] oldD;
   }
   *pXlog = apXlog.get();
   *pDspxx = apDspxx.get();
   *pF2 = apF2.get();
   *pSz = sz;      
}
}
